from __future__ import annotations

from ctypes import (
    CFUNCTYPE,
    POINTER,
    Structure,
    byref,
    c_byte,
    c_char,
    c_char_p,
    c_int,
    c_int16,
    c_int32,
    c_int64,
    c_long,
    c_short,
    c_size_t,
    c_ubyte,
    c_uint,
    c_uint16,
    c_uint32,
    c_uint64,
    c_ulong,
    c_ushort,
    c_void_p,
)
from typing import Any, Callable, Iterable, NoReturn

import pyglet.lib

from .base import FontException

_libfreetype = pyglet.lib.load_library("freetype")

_font_data = {}


def _get_function(name: str, argtypes: Iterable[Any], rtype: Any) -> Callable:
    try:
        func = getattr(_libfreetype, name)
        func.argtypes = argtypes
        func.restype = rtype
        return func
    except AttributeError as e:
        raise ImportError(e)  # noqa: B904


FT_Byte = c_char
FT_Bytes = POINTER(FT_Byte)
FT_Char = c_byte
FT_Int = c_int
FT_UInt = c_uint
FT_Int16 = c_int16
FT_UInt16 = c_uint16
FT_Int32 = c_int32
FT_UInt32 = c_uint32
FT_Int64 = c_int64
FT_UInt64 = c_uint64
FT_Short = c_short
FT_UShort = c_ushort
FT_Long = c_long
FT_ULong = c_ulong
FT_Bool = c_char
FT_Offset = c_size_t
# FT_PtrDist = ?
FT_String = c_char
FT_String_Ptr = c_char_p
FT_Tag = FT_UInt32
FT_Error = c_int
FT_Fixed = c_long
FT_Pointer = c_void_p
FT_Pos = c_long


class FT_Vector(Structure):
    _fields_ = [
        ("x", FT_Pos),
        ("y", FT_Pos),
    ]


class FT_BBox(Structure):
    _fields_ = [
        ("xMin", FT_Pos),
        ("yMin", FT_Pos),
        ("xMax", FT_Pos),
        ("yMax", FT_Pos),
    ]


class FT_Matrix(Structure):
    _fields_ = [
        ("xx", FT_Fixed),
        ("xy", FT_Fixed),
        ("yx", FT_Fixed),
        ("yy", FT_Fixed),
    ]


FT_FWord = c_short
FT_UFWord = c_ushort
FT_F2Dot14 = c_short


class FT_UnitVector(Structure):
    _fields_ = [
        ("x", FT_F2Dot14),
        ("y", FT_F2Dot14),
    ]


FT_F26Dot6 = c_long


class FT_Data(Structure):
    _fields_ = [
        ("pointer", POINTER(FT_Byte)),
        ("length", FT_Int),
    ]


FT_Generic_Finalizer = CFUNCTYPE(None, (c_void_p))


class FT_Generic(Structure):
    _fields_ = [
        ("data", c_void_p),
        ("finalizer", FT_Generic_Finalizer),
    ]


class FT_Bitmap(Structure):
    _fields_ = [
        ("rows", c_uint),
        ("width", c_uint),
        ("pitch", c_int),
        ("buffer", POINTER(c_ubyte)),
        ("num_grays", c_short),
        ("pixel_mode", c_ubyte),
        ("palette_mode", c_ubyte),
        ("palette", c_void_p),
    ]


FT_PIXEL_MODE_NONE = 0
FT_PIXEL_MODE_MONO = 1
FT_PIXEL_MODE_GRAY = 2
FT_PIXEL_MODE_GRAY2 = 3
FT_PIXEL_MODE_GRAY4 = 4
FT_PIXEL_MODE_LCD = 5
FT_PIXEL_MODE_LCD_V = 6
FT_PIXEL_MODE_BGRA = 7


class FT_LibraryRec(Structure):
    _fields_ = [
        ("dummy", c_int),
    ]

    def __del__(self) -> None:
        global _library  # noqa: PLW0603
        try:
            print("FT_LibraryRec.__del__") 
            FT_Done_FreeType(byref(self))
            _library = None
        except:  # noqa: S110, E722
            pass


FT_Library = POINTER(FT_LibraryRec)


class FT_Bitmap_Size(Structure):
    _fields_ = [
        ("height", c_ushort),
        ("width", c_ushort),
        ("size", c_long),
        ("x_ppem", c_long),
        ("y_ppem", c_long),
    ]


class FT_Glyph_Metrics(Structure):
    _fields_ = [
        ("width", FT_Pos),
        ("height", FT_Pos),

        ("horiBearingX", FT_Pos),
        ("horiBearingY", FT_Pos),
        ("horiAdvance", FT_Pos),

        ("vertBearingX", FT_Pos),
        ("vertBearingY", FT_Pos),
        ("vertAdvance", FT_Pos),
    ]

    def dump(self) -> None:
        for (name, _) in self._fields_:
            print("FT_Glyph_Metrics", name, repr(getattr(self, name))) 


FT_Glyph_Format = c_ulong


def FT_IMAGE_TAG(tag: str) -> int:  # noqa: N802
    return (ord(tag[0]) << 24) | (ord(tag[1]) << 16) | (ord(tag[2]) << 8) | ord(tag[3])


FT_GLYPH_FORMAT_NONE = 0
FT_GLYPH_FORMAT_COMPOSITE = FT_IMAGE_TAG("comp")
FT_GLYPH_FORMAT_BITMAP = FT_IMAGE_TAG("bits")
FT_GLYPH_FORMAT_OUTLINE = FT_IMAGE_TAG("outl")
FT_GLYPH_FORMAT_PLOTTER = FT_IMAGE_TAG("plot")


class FT_Outline(Structure):
    _fields_ = [
        ("n_contours", c_short),  # number of contours in glyph
        ("n_points", c_short),  # number of points in the glyph
        ("points", POINTER(FT_Vector)),  # the outline's points
        ("tags", c_char_p),  # the points flags
        ("contours", POINTER(c_short)),  # the contour end points
        ("flags", c_int),  # outline masks
    ]


FT_SubGlyph = c_void_p


class FT_GlyphSlotRec(Structure):
    _fields_ = [
        ("library", FT_Library),
        ("face", c_void_p),
        ("next", c_void_p),
        ("reserved", FT_UInt),
        ("generic", FT_Generic),

        ("metrics", FT_Glyph_Metrics),
        ("linearHoriAdvance", FT_Fixed),
        ("linearVertAdvance", FT_Fixed),
        ("advance", FT_Vector),

        ("format", FT_Glyph_Format),

        ("bitmap", FT_Bitmap),
        ("bitmap_left", FT_Int),
        ("bitmap_top", FT_Int),

        ("outline", FT_Outline),
        ("num_subglyphs", FT_UInt),
        ("subglyphs", FT_SubGlyph),

        ("control_data", c_void_p),
        ("control_len", c_long),

        ("lsb_delta", FT_Pos),
        ("rsb_delta", FT_Pos),

        ("other", c_void_p),

        ("internal", c_void_p),
    ]


FT_GlyphSlot = POINTER(FT_GlyphSlotRec)


class FT_Size_Metrics(Structure):
    _fields_ = [
        ("x_ppem", FT_UShort),  # horizontal pixels per EM
        ("y_ppem", FT_UShort),  # vertical pixels per EM

        ("x_scale", FT_Fixed),  # two scales used to convert font units
        ("y_scale", FT_Fixed),  # to 26.6 frac. pixel coordinates

        ("ascender", FT_Pos),  # ascender in 26.6 frac. pixels
        ("descender", FT_Pos),  # descender in 26.6 frac. pixels
        ("height", FT_Pos),  # text height in 26.6 frac. pixels
        ("max_advance", FT_Pos),  # max horizontal advance, in 26.6 pixels
    ]


class FT_SizeRec(Structure):
    _fields_ = [
        ("face", c_void_p),
        ("generic", FT_Generic),
        ("metrics", FT_Size_Metrics),
        ("internal", c_void_p),
    ]


FT_Size = POINTER(FT_SizeRec)


class FT_FaceRec(Structure):
    _fields_ = [
        ("num_faces", FT_Long),
        ("face_index", FT_Long),

        ("face_flags", FT_Long),
        ("style_flags", FT_Long),

        ("num_glyphs", FT_Long),

        ("family_name", FT_String_Ptr),
        ("style_name", FT_String_Ptr),

        ("num_fixed_sizes", FT_Int),
        ("available_sizes", POINTER(FT_Bitmap_Size)),

        ("num_charmaps", FT_Int),
        ("charmaps", c_void_p),

        ("generic", FT_Generic),

        ("bbox", FT_BBox),

        ("units_per_EM", FT_UShort),
        ("ascender", FT_Short),
        ("descender", FT_Short),
        ("height", FT_Short),

        ("max_advance_width", FT_Short),
        ("max_advance_height", FT_Short),

        ("underline_position", FT_Short),
        ("underline_thickness", FT_Short),

        ("glyph", FT_GlyphSlot),
        ("size", FT_Size),
        ("charmap", c_void_p),

        ("driver", c_void_p),
        ("memory", c_void_p),
        ("stream", c_void_p),

        ("sizes_list", c_void_p),

        ("autohint", FT_Generic),
        ("extensions", c_void_p),
        ("internal", c_void_p),
    ]

    def dump(self) -> None:
        for (name, _) in self._fields_:
            print("FT_FaceRec", name, repr(getattr(self, name))) 

    def has_kerning(self) -> bool:
        return self.face_flags & FT_FACE_FLAG_KERNING


FT_Face = POINTER(FT_FaceRec)

# face_flags values
FT_FACE_FLAG_SCALABLE = 1 << 0
FT_FACE_FLAG_FIXED_SIZES = 1 << 1
FT_FACE_FLAG_FIXED_WIDTH = 1 << 2
FT_FACE_FLAG_SFNT = 1 << 3
FT_FACE_FLAG_HORIZONTAL = 1 << 4
FT_FACE_FLAG_VERTICAL = 1 << 5
FT_FACE_FLAG_KERNING = 1 << 6
FT_FACE_FLAG_FAST_GLYPHS = 1 << 7
FT_FACE_FLAG_MULTIPLE_MASTERS = 1 << 8
FT_FACE_FLAG_GLYPH_NAMES = 1 << 9
FT_FACE_FLAG_EXTERNAL_STREAM = 1 << 10
FT_FACE_FLAG_HINTER = 1 << 11

FT_STYLE_FLAG_ITALIC = 1
FT_STYLE_FLAG_BOLD = 2

(FT_RENDER_MODE_NORMAL,
 FT_RENDER_MODE_LIGHT,
 FT_RENDER_MODE_MONO,
 FT_RENDER_MODE_LCD,
 FT_RENDER_MODE_LCD_V) = range(5)


def FT_LOAD_TARGET_(x: int) -> int:  # noqa: N802
    return (x & 15) << 16


FT_LOAD_TARGET_NORMAL = FT_LOAD_TARGET_(FT_RENDER_MODE_NORMAL)
FT_LOAD_TARGET_LIGHT = FT_LOAD_TARGET_(FT_RENDER_MODE_LIGHT)
FT_LOAD_TARGET_MONO = FT_LOAD_TARGET_(FT_RENDER_MODE_MONO)
FT_LOAD_TARGET_LCD = FT_LOAD_TARGET_(FT_RENDER_MODE_LCD)
FT_LOAD_TARGET_LCD_V = FT_LOAD_TARGET_(FT_RENDER_MODE_LCD_V)


def f16p16_to_float(value: float) -> float:
    return float(value) / (1 << 16)


def float_to_f16p16(value: float) -> float:
    return int(value * (1 << 16))


def f26p6_to_float(value: float) -> float:
    return float(value) / (1 << 6)


def float_to_f26p6(value: float) -> float:
    return int(value * (1 << 6))


class FreeTypeError(FontException):
    def __init__(self, message: str | None, errcode: int) -> None:
        self.message = message
        self.errcode = errcode

    def __str__(self) -> str:
        return "{}: {} ({})".format(self.__class__.__name__, self.message,
                                    self._ft_errors.get(self.errcode, "unknown error"))

    @classmethod
    def check_and_raise_on_error(cls: type[FreeTypeError], errcode: int) -> NoReturn:
        if errcode != 0:
            raise cls(None, errcode)

    _ft_errors = {
        0x00: "no error",
        0x01: "cannot open resource",
        0x02: "unknown file format",
        0x03: "broken file",
        0x04: "invalid FreeType version",
        0x05: "module version is too low",
        0x06: "invalid argument",
        0x07: "unimplemented feature",
        0x08: "broken table",
        0x09: "broken offset within table",
        0x10: "invalid glyph index",
        0x11: "invalid character code",
        0x12: "unsupported glyph image format",
        0x13: "cannot render this glyph format",
        0x14: "invalid outline",
        0x15: "invalid composite glyph",
        0x16: "too many hints",
        0x17: "invalid pixel size",
        0x20: "invalid object handle",
        0x21: "invalid library handle",
        0x22: "invalid module handle",
        0x23: "invalid face handle",
        0x24: "invalid size handle",
        0x25: "invalid glyph slot handle",
        0x26: "invalid charmap handle",
        0x27: "invalid cache manager handle",
        0x28: "invalid stream handle",
        0x30: "too many modules",
        0x31: "too many extensions",
        0x40: "out of memory",
        0x41: "unlisted object",
        0x51: "cannot open stream",
        0x52: "invalid stream seek",
        0x53: "invalid stream skip",
        0x54: "invalid stream read",
        0x55: "invalid stream operation",
        0x56: "invalid frame operation",
        0x57: "nested frame access",
        0x58: "invalid frame read",
        0x60: "raster uninitialized",
        0x61: "raster corrupted",
        0x62: "raster overflow",
        0x63: "negative height while rastering",
        0x70: "too many registered caches",
        0x80: "invalid opcode",
        0x81: "too few arguments",
        0x82: "stack overflow",
        0x83: "code overflow",
        0x84: "bad argument",
        0x85: "division by zero",
        0x86: "invalid reference",
        0x87: "found debug opcode",
        0x88: "found ENDF opcode in execution stream",
        0x89: "nested DEFS",
        0x8A: "invalid code range",
        0x8B: "execution context too long",
        0x8C: "too many function definitions",
        0x8D: "too many instruction definitions",
        0x8E: "SFNT font table missing",
        0x8F: "horizontal header (hhea, table missing",
        0x90: "locations (loca, table missing",
        0x91: "name table missing",
        0x92: "character map (cmap, table missing",
        0x93: "horizontal metrics (hmtx, table missing",
        0x94: "PostScript (post, table missing",
        0x95: "invalid horizontal metrics",
        0x96: "invalid character map (cmap, format",
        0x97: "invalid ppem value",
        0x98: "invalid vertical metrics",
        0x99: "could not find context",
        0x9A: "invalid PostScript (post, table format",
        0x9B: "invalid PostScript (post, table",
        0xA0: "opcode syntax error",
        0xA1: "argument stack underflow",
        0xA2: "ignore",
        0xB0: "`STARTFONT' field missing",
        0xB1: "`FONT' field missing",
        0xB2: "`SIZE' field missing",
        0xB3: "`CHARS' field missing",
        0xB4: "`STARTCHAR' field missing",
        0xB5: "`ENCODING' field missing",
        0xB6: "`BBX' field missing",
        0xB7: "`BBX' too big",
    }


def _get_function_with_error_handling(name: str, argtypes: Iterable[Any], rtype: Any) -> Callable:
    func = _get_function(name, argtypes, rtype)

    def _error_handling(*args, **kwargs) -> None:  # noqa: ANN002, ANN003
        err = func(*args, **kwargs)
        FreeTypeError.check_and_raise_on_error(err)

    return _error_handling


FT_LOAD_RENDER = 0x4

FT_Init_FreeType = _get_function_with_error_handling("FT_Init_FreeType",
                                                     [POINTER(FT_Library)], FT_Error)
FT_Done_FreeType = _get_function_with_error_handling("FT_Done_FreeType",
                                                     [FT_Library], FT_Error)

FT_New_Face = _get_function_with_error_handling("FT_New_Face",
                                                [FT_Library, c_char_p, FT_Long, POINTER(FT_Face)], FT_Error)
FT_Done_Face = _get_function_with_error_handling("FT_Done_Face",
                                                 [FT_Face], FT_Error)
FT_Reference_Face = _get_function_with_error_handling("FT_Reference_Face",
                                                      [FT_Face], FT_Error)
FT_New_Memory_Face = _get_function_with_error_handling("FT_New_Memory_Face",
                                                       [FT_Library, POINTER(FT_Byte), FT_Long, FT_Long,
                                                        POINTER(FT_Face)], FT_Error)

FT_Set_Char_Size = _get_function_with_error_handling("FT_Set_Char_Size",
                                                     [FT_Face, FT_F26Dot6, FT_F26Dot6, FT_UInt, FT_UInt], FT_Error)
FT_Set_Pixel_Sizes = _get_function_with_error_handling("FT_Set_Pixel_Sizes",
                                                       [FT_Face, FT_UInt, FT_UInt], FT_Error)
FT_Load_Glyph = _get_function_with_error_handling("FT_Load_Glyph",
                                                  [FT_Face, FT_UInt, FT_Int32], FT_Error)
FT_Get_Char_Index = _get_function_with_error_handling("FT_Get_Char_Index",
                                                      [FT_Face, FT_ULong], FT_Error)
FT_Load_Char = _get_function_with_error_handling("FT_Load_Char",
                                                 [FT_Face, FT_ULong, FT_Int32], FT_Error)
FT_Get_Kerning = _get_function_with_error_handling("FT_Get_Kerning",
                                                   [FT_Face, FT_UInt, FT_UInt, FT_UInt, POINTER(FT_Vector)], FT_Error)


# SFNT interface

class FT_SfntName(Structure):
    _fields_ = [
        ("platform_id", FT_UShort),
        ("encoding_id", FT_UShort),
        ("language_id", FT_UShort),
        ("name_id", FT_UShort),
        ("string", POINTER(FT_Byte)),
        ("string_len", FT_UInt),
    ]


FT_Get_Sfnt_Name_Count = _get_function("FT_Get_Sfnt_Name_Count",
                                       [FT_Face], FT_UInt)
FT_Get_Sfnt_Name = _get_function_with_error_handling("FT_Get_Sfnt_Name",
                                                     [FT_Face, FT_UInt, POINTER(FT_SfntName)], FT_Error)

TT_PLATFORM_MICROSOFT = 3
TT_MS_ID_UNICODE_CS = 1
TT_NAME_ID_COPYRIGHT = 0
TT_NAME_ID_FONT_FAMILY = 1
TT_NAME_ID_FONT_SUBFAMILY = 2
TT_NAME_ID_UNIQUE_ID = 3
TT_NAME_ID_FULL_NAME = 4
TT_NAME_ID_VERSION_STRING = 5
TT_NAME_ID_PS_NAME = 6
TT_NAME_ID_TRADEMARK = 7
TT_NAME_ID_MANUFACTURER = 8
TT_NAME_ID_DESIGNER = 9
TT_NAME_ID_DESCRIPTION = 10
TT_NAME_ID_VENDOR_URL = 11
TT_NAME_ID_DESIGNER_URL = 12
TT_NAME_ID_LICENSE = 13
TT_NAME_ID_LICENSE_URL = 14
TT_NAME_ID_PREFERRED_FAMILY = 16
TT_NAME_ID_PREFERRED_SUBFAMILY = 17
TT_NAME_ID_MAC_FULL_NAME = 18
TT_NAME_ID_CID_FINDFONT_NAME = 20

_library = None


def ft_get_library() -> FT_Library:
    global _library  # noqa: PLW0603
    if not _library:
        _library = FT_Library()
        error = FT_Init_FreeType(byref(_library))
        if error:
            msg = "an error occurred during library initialization"
            raise FontException(msg, error)
    return _library



"""This module defines the usage and creation of user defined fonts in Pyglet.

Previously, pyglet only supported font renderers that are built into the operating system, such as
``FreeType``, ``DirectWrite``, or ``Quartz``. However, there are situations in which a user may not want or need all the
features a font can provide. They just need to put characters in a particular order without the hassle of exporting
into a separate file.

The :py:class:`~pyglet.font.user.UserDefinedMappingFont` is provided for most use cases, which will allow you to
make an internal font that can be used where a ``font_name`` is required to identify a font.

A user defined font is also identified by its name. The name you choose should be unique to ensure it will not conflict
with a system font. For example, do not use `Arial`, as that will collide with Windows systems.

With :py:class:`~pyglet.font.user.UserDefinedMappingFont` you can pass a mapping of characters that point to your
:py:class:`~pyglet.image.ImageData`.

.. code-block:: python

    mappings={'c': my_image_data, 'b': my_image_data, 'a': my_image_data}

For more custom behavior, a dict-like object can be used, such as a class.

.. code-block:: python

    class MyCustomMapping:
        def get(self, char: str) -> ImageData | None:
            # return ImageData if a character is found.
            # return None if no character is found

    mappings = MyCustomMapping()

Once your font is created, you also must register it within pyglet to use it. This can be done through the
 :py:func:`~pyglet.font.add_user_font` function.

When you register a user defined font, only those parameters will used to identify the font. If you have a font, but
want to have a ``italic`` enabled version. You must make a new instance of your font, but with the ``italic``
parameter set as ``True``. Same applies to the ``size`` parameter. The ``weight`` parameter can also be provided as
a string.

Scaling
=======
By default, user font's will not be scaled. In most use cases, you have a single font at a specific size that you
want to use.

There are cases where a user may want to scale their font to be used at any size. We provide the following function:
:py:func:`~pyglet.font.user.get_scaled_user_font`. By providing the user defined font instance, and a new size, you will
get back a new font instance that is scaled to the new size. This new instance must also be registered the same way as
the base font.

When specifying the ``size`` parameter, that value is used to determine the ratio of scaling between the new size. So
if your base font is a size of 12, creating a scaled version at size 24 will be double the size of the base.

.. warning::

    The ``PIL`` library is a required dependency to use the scaling functionality.

.. versionadded:: 2.0.15
"""
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Protocol

import pyglet
from pyglet.font import base

SCALING_ENABLED = False
try:
    from PIL import Image
    from PIL.Image import Resampling

    SCALING_ENABLED = True
except ImportError:
    pass

if TYPE_CHECKING:
    from pyglet.font.base import Glyph
    from pyglet.image import ImageData


class UserDefinedGlyphRenderer(base.GlyphRenderer):
    def __init__(self, font: UserDefinedFontBase) -> None:
        super().__init__(font)
        self._font = font

    def render(self, image_data: ImageData) -> Glyph:
        if self._font._scaling:  # noqa: SLF001
            image_original = Image.frombytes("RGBA", (image_data.width, image_data.height),
                                             image_data.get_image_data().get_data("RGBA"))
            scale_ratio = self._font.size / self._font._base_size
            image_resized = image_original.resize((int(image_data.width * scale_ratio),
                                                   int(image_data.height * scale_ratio)), Resampling.NEAREST)
            new_image = pyglet.image.ImageData(image_resized.width, image_resized.height,
                                               "RGBA", image_resized.tobytes())
            glyph = self._font.create_glyph(new_image)
            glyph.set_bearings(-self._font.descent, 0, image_resized.width)
        else:
            glyph = self._font.create_glyph(image_data)
            glyph.set_bearings(-self._font.descent, 0, image_data.width)
        return glyph


class UserDefinedFontBase(base.Font):
    """Used as a base for all user defined fonts.

    .. versionadded:: 2.0.15
    """
    glyph_renderer_class: ClassVar[type[base.GlyphRenderer]] = UserDefinedGlyphRenderer

    def __init__(
            self, name: str, default_char: str, size: int, ascent: int | None = None, descent: int | None = None,
            weight: str = "normal", italic: bool = False, stretch: bool = False, dpi: int = 96, locale: str | None = None,
    ) -> None:
        """Initialize a user defined font.

        Args:
            name:
                Name of the font. Used to identify the font. Must be unique to ensure it does not
                collide with any system fonts.
            default_char:
                If a character in a string is not found in the font, it will use this as fallback.
            size:
                Font size, usually in pixels.
            ascent:
                Maximum ascent above the baseline, in pixels. If None, the image height is used.
            descent:
                Maximum descent below the baseline, in pixels. Usually negative.
            weight:
                The font weight, as a string. Defaults to "normal".
            italic:
                If True, this font will be used when ``italic`` is enabled for the font name.
            stretch:
                If True, this font will be used when ``stretch`` is enabled for the font name.
            dpi:
                The assumed resolution of the display device, for the purposes of determining the pixel size of the
                font. Use a default of 96 for standard sizing.
            locale:
                Used to specify the locale of this font.
        """
        super().__init__()
        self._name = name
        self.default_char = default_char
        self.ascent = ascent
        self.descent = descent
        self.size = size
        self.weight = weight
        self.italic = italic
        self.stretch = stretch
        self.dpi = dpi
        self.locale = locale

        self._base_size = 0
        self._scaling = False

    @property
    def name(self) -> str:
        return self._name

    def enable_scaling(self, base_size: int) -> None:
        if not SCALING_ENABLED:
            msg = "PIL is not installed. User Font Scaling requires PIL."
            raise Exception(msg)

        self._base_size = base_size
        self._scaling = True


class UserDefinedFontException(Exception):  # noqa: N818
    """An exception related to user font creation."""


class DictLikeObject(Protocol):
    def get(self, char: str) -> ImageData | None:
        pass


class UserDefinedMappingFont(UserDefinedFontBase):
    """The class allows the creation of user defined fonts from a set of mappings.
    
    .. versionadded:: 2.0.15
    """

    def __init__(self, name: str, default_char: str, size: int, mappings: DictLikeObject,
            ascent: int | None = None, descent: int | None = None, weight: str = "normal", italic: bool = False,
            stretch: bool = False, dpi: int = 96, locale: str | None = None) -> None:
        """Initialize the default parameters of your font.

        Args:
            name:
                Name of the font. Must be unique to ensure it does not collide with any system fonts.
            default_char:
                If a character in a string is not found in the font, it will use this as fallback.
            size:
                Font size. Should be in pixels. This value will affect scaling if enabled.
            mappings:
                A dict or dict-like object with a ``get`` function.
                The ``get`` function must take a string character, and output :py:class:`~pyglet.iamge.ImageData` if
                found. It also must return ``None`` if no character is found.
            ascent:
                Maximum ascent above the baseline, in pixels. If None, the image height is used.
            descent:
                Maximum descent below the baseline, in pixels. Usually negative.
            weight:
                The font weight, as a string. Defaults to "normal".
            italic:
                If ``True``, this font will be used when ``italic`` is enabled for the font name.
            stretch:
                If ``True``, this font will be used when ``stretch`` is enabled for the font name.
            dpi:
                The assumed resolution of the display device, for the purposes of determining the pixel size of the
                font. Use a default of 96 for standard sizing.
            locale:
                Used to specify the locale of this font.
        """
        self.mappings = mappings
        default_image = self.mappings.get(default_char)
        if not default_image:
            msg = f"Default character '{default_char}' must exist within your mappings."
            raise UserDefinedFontException(msg)

        if ascent is None or descent is None:
            if ascent is None:
                ascent = default_image.height
            if descent is None:
                descent = 0

        super().__init__(name, default_char, size, ascent, descent, weight, italic, stretch, dpi, locale)

    def enable_scaling(self, base_size: int) -> None:
        """Enables scaling the font size.

        Args:
            base_size:
                The base size is used to calculate the ratio between new sizes and the original.
        """
        super().enable_scaling(base_size)
        glyphs = self.get_glyphs(self.default_char)
        self.ascent = glyphs[0].height
        self.descent = 0

    def get_glyphs(self, text: str) -> list[Glyph]:
        """Create and return a list of Glyphs for `text`.

        If any characters do not have a known glyph representation in this font, a substitution will be made with
        the ``default_char``.
        """
        glyph_renderer = None
        glyphs = []  # glyphs that are committed.
        for c in base.get_grapheme_clusters(text):
            # Get the glyph for 'c'.  Hide tabs (Windows and Linux render
            # boxes)
            if c == "\t":
                c = " "
            if c not in self.glyphs:
                if not glyph_renderer:
                    glyph_renderer = self.glyph_renderer_class(self)

                image_data = self.mappings.get(c)
                if not image_data:
                    c = self.default_char
                else:
                    self.glyphs[c] = glyph_renderer.render(image_data)
            glyphs.append(self.glyphs[c])
        return glyphs


def get_scaled_user_font(font_base: UserDefinedMappingFont, size: int) -> UserDefinedMappingFont:
    """This function will return a new font instance which can scale it's size based off the original base font.

    .. note:: The scaling functionality requires the PIL library to be installed.

    .. versionadded:: 2.0.15

    Args:
        font_base:
            The base font object to create a new size from.
        size:
            The new font size. This will be scaled based on the ratio between the base size and the new size.
    """
    new_font = UserDefinedMappingFont(font_base.name, font_base.default_char, size, font_base.mappings,
                                      font_base.ascent, font_base.descent, font_base.weight, font_base.italic,
                                      font_base.stretch, font_base.dpi, font_base.locale)

    new_font.enable_scaling(font_base.size)
    return new_font


__all__ = (
    "UserDefinedFontBase",
    "UserDefinedFontException",
    "UserDefinedMappingFont",
    "get_scaled_user_font",
)



from __future__ import annotations

import warnings
from ctypes import POINTER, byref, c_ubyte, cast, memmove
from typing import NamedTuple

import pyglet
from pyglet import image
from pyglet.font import base
from pyglet.font.fontconfig import get_fontconfig
from pyglet.font.freetype_lib import (
    FT_LOAD_RENDER,
    FT_PIXEL_MODE_GRAY,
    FT_PIXEL_MODE_MONO,
    FT_STYLE_FLAG_BOLD,
    FT_STYLE_FLAG_ITALIC,
    FreeTypeError,
    FT_Byte,
    FT_Done_Face,
    FT_Face,
    FT_GlyphSlot,
    FT_Load_Glyph,
    FT_New_Face,
    FT_New_Memory_Face,
    FT_Reference_Face,
    FT_Set_Char_Size,
    f26p6_to_float,
    float_to_f26p6,
    ft_get_library,
)
from pyglet.util import asbytes, asstr


class FreeTypeGlyphRenderer(base.GlyphRenderer):
    def __init__(self, font: FreeTypeFont) -> None:
        super().__init__(font)
        self.font = font

        self._glyph_slot = None
        self._bitmap = None

        self._width = None
        self._height = None
        self._mode = None
        self._pitch = None

        self._baseline = None
        self._lsb = None
        self._advance_x = None

        self._data = None

    def _get_glyph(self, character: str) -> None:
        assert self.font
        assert len(character) == 1

        self._glyph_slot = self.font.get_glyph_slot(character)
        self._bitmap = self._glyph_slot.bitmap

    def _get_glyph_metrics(self) -> None:
        self._width = self._glyph_slot.bitmap.width
        self._height = self._glyph_slot.bitmap.rows
        self._mode = self._glyph_slot.bitmap.pixel_mode
        self._pitch = self._glyph_slot.bitmap.pitch

        self._baseline = self._height - self._glyph_slot.bitmap_top
        self._lsb = self._glyph_slot.bitmap_left
        self._advance_x = int(f26p6_to_float(self._glyph_slot.advance.x))

    def _get_bitmap_data(self) -> None:
        if self._mode == FT_PIXEL_MODE_MONO:
            # BCF fonts always render to 1 bit mono, regardless of render
            # flags. (freetype 2.3.5)
            self._convert_mono_to_gray_bitmap()
        elif self._mode == FT_PIXEL_MODE_GRAY:
            # Usual case
            assert self._glyph_slot.bitmap.num_grays == 256
            self._data = self._glyph_slot.bitmap.buffer
        else:
            msg = "Unsupported render mode for this glyph"
            raise base.FontException(msg)

    def _convert_mono_to_gray_bitmap(self) -> None:
        bitmap_data = cast(self._bitmap.buffer,
                           POINTER(c_ubyte * (self._pitch * self._height))).contents
        data = (c_ubyte * (self._pitch * 8 * self._height))()
        data_i = 0
        for byte in bitmap_data:
            # Data is MSB; left-most pixel in a byte has value 128.
            data[data_i + 0] = (byte & 0x80) and 255 or 0
            data[data_i + 1] = (byte & 0x40) and 255 or 0
            data[data_i + 2] = (byte & 0x20) and 255 or 0
            data[data_i + 3] = (byte & 0x10) and 255 or 0
            data[data_i + 4] = (byte & 0x08) and 255 or 0
            data[data_i + 5] = (byte & 0x04) and 255 or 0
            data[data_i + 6] = (byte & 0x02) and 255 or 0
            data[data_i + 7] = (byte & 0x01) and 255 or 0
            data_i += 8
        self._data = data
        self._pitch <<= 3

    def _create_glyph(self) -> base.Glyph:
        # In FT positive pitch means `down` flow, in Pyglet ImageData
        # negative values indicate a top-to-bottom arrangement. So pitch must be inverted.
        # Using negative pitch causes conversions, so much faster to just swap tex_coords
        img = image.ImageData(self._width,
                              self._height,
                              "A",
                              self._data,
                              abs(self._pitch))

        # HACK: Get text working in GLES until image data can be converted properly
        #       GLES don't support coversion during pixel transfer so we have to
        #       force specify the glyph format to be GL_ALPHA. This format is not
        #       supported in 3.3+ core, but are present in ES because of pixel transfer
        #       limitations.
        if pyglet.gl.current_context.get_info().get_opengl_api() == "gles":
            GL_ALPHA = 0x1906
            glyph = self.font.create_glyph(img, fmt=GL_ALPHA)
        else:
            glyph = self.font.create_glyph(img)

        glyph.set_bearings(self._baseline, self._lsb, self._advance_x)
        if self._pitch > 0:
            t = list(glyph.tex_coords)
            glyph.tex_coords = t[9:12] + t[6:9] + t[3:6] + t[:3]

        return glyph

    def render(self, text: str) -> base.Glyph:
        self._get_glyph(text[0])
        self._get_glyph_metrics()
        self._get_bitmap_data()
        return self._create_glyph()


class FreeTypeFontMetrics(NamedTuple):
    ascent: int
    descent: int


class MemoryFaceStore:
    _dict: dict[tuple[str, str, bool], FreeTypeMemoryFace]

    def __init__(self) -> None:
        self._dict = {}

    def add(self, face: FreeTypeMemoryFace) -> None:
        self._dict[face.name.lower(), face.weight, face.italic] = face

    def contains(self, name: str) -> bool:
        lname = name and name.lower() or ""
        return len([name for name, _, _ in self._dict if name == lname]) > 0

    def get(self, name: str, weight: str, italic: bool) -> FreeTypeMemoryFace | None:
        lname = name and name.lower() or ""
        return self._dict.get((lname, weight, italic), None)


class FreeTypeFont(base.Font):
    glyph_renderer_class = FreeTypeGlyphRenderer

    # Map font (name, weight, italic) to FreeTypeMemoryFace
    _memory_faces = MemoryFaceStore()
    face: FreeTypeFace

    def __init__(self, name: str, size: float, weight: str = "normal", italic: bool = False, stretch: bool = False,
                 dpi: int | None = None) -> None:

        if stretch:
            warnings.warn("The current font render does not support stretching.")  # noqa: B028

        super().__init__()
        self._name = name
        self.size = size
        self.weight = weight
        self.italic = italic
        self.dpi = dpi or 96

        self._load_font_face()
        self.metrics = self.face.get_font_metrics(self.size, self.dpi)

    @property
    def name(self) -> str:
        return self.face.family_name

    @property
    def ascent(self) -> int:
        return self.metrics.ascent

    @property
    def descent(self) -> int:
        return self.metrics.descent

    def get_glyph_slot(self, character: str) -> FT_GlyphSlot:
        glyph_index = self.face.get_character_index(character)
        self.face.set_char_size(self.size, self.dpi)
        return self.face.get_glyph_slot(glyph_index)

    def _load_font_face(self) -> None:
        self.face = self._memory_faces.get(self._name, self.weight, self.italic)
        if self.face is None:
            self._load_font_face_from_system()

    def _load_font_face_from_system(self) -> None:
        match = get_fontconfig().find_font(self._name, self.size, self.weight, self.italic)
        if not match:
            msg = f"Could not match font '{self._name}'"
            raise base.FontException(msg)
        self.filename = match.file
        self.face = FreeTypeFace.from_fontconfig(match)

    @classmethod
    def have_font(cls: type[FreeTypeFont], name: str) -> bool:
        if cls._memory_faces.contains(name):
            return True

        return get_fontconfig().have_font(name)

    @classmethod
    def add_font_data(cls: type[FreeTypeFont], data: bytes) -> None:
        face = FreeTypeMemoryFace(data)
        cls._memory_faces.add(face)


class FreeTypeFace:
    """FreeType typographic face object.

    Keeps the reference count to the face at +1 as long as this object exists. If other objects
    want to keep a face without a reference to this object, they should increase the reference
    counter themselves and decrease it again when done.
    """
    _name: str
    ft_face: FT_Face

    def __init__(self, ft_face: FT_Face) -> None:
        assert ft_face is not None
        self.ft_face = ft_face
        self._get_best_name()

    @classmethod
    def from_file(cls: type[FreeTypeFace], file_name: str) -> FreeTypeFace:
        ft_library = ft_get_library()
        ft_face = FT_Face()
        FT_New_Face(ft_library,
                    asbytes(file_name),
                    0,
                    byref(ft_face))
        return cls(ft_face)

    @classmethod
    def from_fontconfig(cls: type[FreeTypeFace], match):
        if match.face is not None:
            FT_Reference_Face(match.face)
            return cls(match.face)
        else:
            if not match.file:
                msg = f'No filename for "{match.name}"'
                raise base.FontException(msg)
            return cls.from_file(match.file)

    @property
    def name(self) -> str:
        return self._name

    @property
    def family_name(self) -> str:
        return asstr(self.ft_face.contents.family_name)

    @property
    def style_flags(self) -> int:
        return self.ft_face.contents.style_flags

    @property
    def bold(self) -> bool:
        return self.style_flags & FT_STYLE_FLAG_BOLD != 0

    @property
    def weight(self) -> str:
        return "normal"

    @property
    def italic(self) -> bool:
        return self.style_flags & FT_STYLE_FLAG_ITALIC != 0

    @property
    def face_flags(self) -> int:
        return self.ft_face.contents.face_flags

    def __del__(self) -> None:
        if self.ft_face is not None:
            FT_Done_Face(self.ft_face)
            self.ft_face = None

    def set_char_size(self, size: float, dpi: int) -> bool:
        face_size = float_to_f26p6(size)
        try:
            FT_Set_Char_Size(self.ft_face,
                             0,
                             face_size,
                             dpi,
                             dpi)
            return True
        except FreeTypeError as e:
            # Error 0x17 indicates invalid pixel size, so font size cannot be changed
            # TODO Warn the user?
            if e.errcode == 0x17:
                return False

            raise

    def get_character_index(self, character: str) -> int:
        return get_fontconfig().char_index(self.ft_face, character)

    def get_glyph_slot(self, glyph_index: int) -> FT_GlyphSlot:
        FT_Load_Glyph(self.ft_face, glyph_index, FT_LOAD_RENDER)
        return self.ft_face.contents.glyph.contents

    def get_font_metrics(self, size: float, dpi: int) -> FreeTypeFontMetrics:
        if self.set_char_size(size, dpi):
            metrics = self.ft_face.contents.size.contents.metrics
            if metrics.ascender == 0 and metrics.descender == 0:
                return self._get_font_metrics_workaround()

            return FreeTypeFontMetrics(ascent=int(f26p6_to_float(metrics.ascender)),
                                       descent=int(f26p6_to_float(metrics.descender)))

        return self._get_font_metrics_workaround()

    def _get_font_metrics_workaround(self) -> FreeTypeFontMetrics:
        # Workaround broken fonts with no metrics.  Has been observed with
        # courR12-ISO8859-1.pcf.gz: "Courier" "Regular"
        #
        # None of the metrics fields are filled in, so render a glyph and
        # grab its height as the ascent, and make up an arbitrary
        # descent.
        i = self.get_character_index("X")
        self.get_glyph_slot(i)
        ascent = self.ft_face.contents.available_sizes.contents.height
        return FreeTypeFontMetrics(ascent=ascent,
                                   descent=-ascent // 4)  # arbitrary.

    def _get_best_name(self) -> None:
        self._name = asstr(self.ft_face.contents.family_name)
        self._get_font_family_from_ttf()

    def _get_font_family_from_ttf(self) -> None:
        # Replace Freetype's generic family name with TTF/OpenType specific
        # name if we can find one; there are some instances where Freetype
        # gets it wrong.

        return  # FIXME: This is broken

        if self.face_flags & FT_FACE_FLAG_SFNT:
            name = FT_SfntName()
            for i in range(FT_Get_Sfnt_Name_Count(self.ft_face)):
                try:
                    FT_Get_Sfnt_Name(self.ft_face, i, name)
                    if not (name.platform_id == TT_PLATFORM_MICROSOFT and
                            name.encoding_id == TT_MS_ID_UNICODE_CS):
                        continue
                    # name.string is not 0 terminated! use name.string_len
                    self._name = name.string.decode("utf-16be", "ignore")
                except:
                    continue


class FreeTypeMemoryFace(FreeTypeFace):
    def __init__(self, data: bytes) -> None:
        self._copy_font_data(data)
        super().__init__(self._create_font_face())

    def _copy_font_data(self, data: bytes) -> None:
        self.font_data = (FT_Byte * len(data))()
        memmove(self.font_data, data, len(data))

    def _create_font_face(self) -> FT_Face:
        ft_library = ft_get_library()
        ft_face = FT_Face()
        FT_New_Memory_Face(ft_library,
                           self.font_data,
                           len(self.font_data),
                           0,
                           byref(ft_face))
        return ft_face



from .base import Display, Screen, ScreenMode, Canvas

from pyglet.libs.win32 import _user32, _shcore, _gdi32
from pyglet.libs.win32.constants import (
    CDS_FULLSCREEN,
    DISP_CHANGE_SUCCESSFUL,
    ENUM_CURRENT_SETTINGS,
    WINDOWS_8_1_OR_GREATER,
    WINDOWS_VISTA_OR_GREATER,
    WINDOWS_10_CREATORS_UPDATE_OR_GREATER,
    USER_DEFAULT_SCREEN_DPI,
    LOGPIXELSX,
    LOGPIXELSY,

)
from pyglet.libs.win32.types import (
    DEVMODE,
    DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2,
    MONITORINFOEX,
    MONITORENUMPROC,
    PROCESS_PER_MONITOR_DPI_AWARE,
    UINT,
    sizeof,
    byref
)
from pyglet.libs.win32.context_managers import device_context


def set_dpi_awareness():
    """
       Setting DPI varies per Windows version.
       Note: DPI awareness needs to be set before Window, Display, or Screens are initialized.
    """
    if WINDOWS_10_CREATORS_UPDATE_OR_GREATER:
        _user32.SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2)
    elif WINDOWS_8_1_OR_GREATER:  # 8.1 and above allows per monitor DPI.
        _shcore.SetProcessDpiAwareness(PROCESS_PER_MONITOR_DPI_AWARE)
    elif WINDOWS_VISTA_OR_GREATER:  # Only System wide DPI
        _user32.SetProcessDPIAware()


set_dpi_awareness()


class Win32Display(Display):
    def get_screens(self):
        screens = []

        def enum_proc(hMonitor, hdcMonitor, lprcMonitor, dwData):
            r = lprcMonitor.contents
            width = r.right - r.left
            height = r.bottom - r.top
            screens.append(
                Win32Screen(self, hMonitor, r.left, r.top, width, height))
            return True

        enum_proc_ptr = MONITORENUMPROC(enum_proc)
        _user32.EnumDisplayMonitors(None, None, enum_proc_ptr, 0)
        return screens


class Win32Screen(Screen):
    _initial_mode = None

    def __init__(self, display, handle, x, y, width, height):
        super(Win32Screen, self).__init__(display, x, y, width, height)
        self._handle = handle

    def get_matching_configs(self, template):
        with device_context(None) as hdc:
            canvas = Win32Canvas(self.display, 0, hdc)
            configs = template.match(canvas)
            # XXX deprecate config's being screen-specific
            for config in configs:
                config.screen = self

        return configs

    def get_device_name(self):
        info = MONITORINFOEX()
        info.cbSize = sizeof(MONITORINFOEX)
        _user32.GetMonitorInfoW(self._handle, byref(info))
        return info.szDevice

    def get_dpi(self):
        if WINDOWS_8_1_OR_GREATER:
            xdpi = UINT()
            ydpi = UINT()
            _shcore.GetDpiForMonitor(self._handle, 0, byref(xdpi), byref(ydpi))
            xdpi, ydpi = xdpi.value, ydpi.value
        else:
            dc = _user32.GetDC(None)
            xdpi = _gdi32.GetDeviceCaps(dc, LOGPIXELSX)
            ydpi = _gdi32.GetDeviceCaps(dc, LOGPIXELSY)
            _user32.ReleaseDC(0, dc)

        return xdpi

    def get_scale(self):
        xdpi = self.get_dpi()
        return xdpi / USER_DEFAULT_SCREEN_DPI

    def get_modes(self):
        device_name = self.get_device_name()
        i = 0
        modes = []
        while True:
            mode = DEVMODE()
            mode.dmSize = sizeof(DEVMODE)
            r = _user32.EnumDisplaySettingsW(device_name, i, byref(mode))
            if not r:
                break

            modes.append(Win32ScreenMode(self, mode))
            i += 1

        return modes

    def get_mode(self):
        mode = DEVMODE()
        mode.dmSize = sizeof(DEVMODE)
        _user32.EnumDisplaySettingsW(self.get_device_name(),
                                     ENUM_CURRENT_SETTINGS,
                                     byref(mode))
        return Win32ScreenMode(self, mode)

    def set_mode(self, mode):
        assert mode.screen is self

        if not self._initial_mode:
            self._initial_mode = self.get_mode()
        r = _user32.ChangeDisplaySettingsExW(self.get_device_name(),
                                             byref(mode._mode),
                                             None,
                                             CDS_FULLSCREEN,
                                             None)
        if r == DISP_CHANGE_SUCCESSFUL:
            self.width = mode.width
            self.height = mode.height

    def restore_mode(self):
        if self._initial_mode:
            self.set_mode(self._initial_mode)


class Win32ScreenMode(ScreenMode):
    def __init__(self, screen, mode):
        super(Win32ScreenMode, self).__init__(screen)
        self._mode = mode
        self.width = mode.dmPelsWidth
        self.height = mode.dmPelsHeight
        self.depth = mode.dmBitsPerPel
        self.rate = mode.dmDisplayFrequency
        self.scaling = mode.dmDisplayFixedOutput

    def __repr__(self):
        return f'{self.__class__.__name__}(width={self.width!r}, height={self.height!r}, depth={self.depth!r}, rate={self.rate}, scaling={self.scaling})'

class Win32Canvas(Canvas):
    def __init__(self, display, hwnd, hdc):
        super(Win32Canvas, self).__init__(display)
        self.hwnd = hwnd
        self.hdc = hdc



"""Generate gl.py and gl_compat.py
See : tools/requirements.txt.

We are using the opengl-registry project to extract this information from
https://raw.githubusercontent.com/KhronosGroup/OpenGL-Registry/master/xml/gl.xml

A local version gl.xml can also be used.

Usage:

# Fetch gl.xml from Khronos Github repo
python gengl.py
python gengl.py --source url

# Use local gl.xml
python gengl.py --source local
"""  # noqa: D205
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path

from opengl_registry import Registry, RegistryReader

REPO_ROOT = Path(__file__).parent.parent.resolve()
DEST_PATH = REPO_ROOT / "pyglet" / "gl"


def main() -> None:  # noqa: D103
    values = parse_args(sys.argv[1:])
    if values.source == "url":
        # Fetch gl.xml from Khronos Github repo
        reader = RegistryReader.from_url()
    else:
        # Use the local gl.xml file
        reader = RegistryReader.from_file(Path(REPO_ROOT / "tools" / "gl.xml"))

    registry = reader.read()

    # OpenGL extensions we want to include
    extensions = [
        "GL_ARB_multisample",
        "EXT_texture_compression_s3tc",  # For pyglet.image.codecs.dds
        "GL_EXT_framebuffer_object",  # Needed for GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS
        "GL_ARB_bindless_texture",
        "GL_ARB_gpu_shader_int64",
        "GL_NV_mesh_shader",
    ]

    core_profile = registry.get_profile(
        api="gl",
        profile="core",
        version="4.6",
        extensions=extensions,
    )
    compat_profile = registry.get_profile(
        api="gl",
        profile="compat",
        version="4.6",
        extensions=extensions,
    )
    # es_profile = registry.get_profile(
    #     api="gles2",
    #     profile="core",
    #     version="3.2",
    #     extensions=extensions,
    # )

    core_writer = PygletGLWriter(registry=core_profile, out_file=DEST_PATH / "gl.py")
    core_writer.run()
    compat_writer = PygletGLWriter(registry=compat_profile, out_file=DEST_PATH / "gl_compat.py")
    compat_writer.run()
    # es_writer = PygletGLWriter(registry=es_profile, out_file=DEST_PATH / "gl_es.py")
    # es_writer.run()


def parse_args(args: str) -> Namespace:  # noqa: D103
    parser = ArgumentParser()
    parser.add_argument("--source", choices=["local", "url"], default="url")
    return parser.parse_args(args)


class PygletGLWriter:
    """Writes gl.py, gl_compat.py, and gl_es.py."""

    # All gl types manually matched to ctypes.
    # Inspect registry.types
    types = {  # noqa: RUF012
        "GLenum": "c_uint",
        "GLboolean": "c_ubyte",
        "GLbitfield": "c_uint",
        "GLvoid": "None",
        "GLbyte": "c_char",
        "GLubyte": "c_ubyte",
        "GLshort": "c_short",
        "GLushort": "c_ushort",
        "GLint": "c_int",
        "GLuint": "c_uint",
        "GLclampx": "c_uint",
        "GLsizei": "c_int",
        "GLfloat": "c_float",
        "GLclampf": "c_float",
        "GLdouble": "c_double",
        "GLclampd": "c_double",
        "GLchar": "c_char",
        "GLintptr": "c_ptrdiff_t",
        "GLsizeiptr": "c_ptrdiff_t",
        "GLint64": "c_int64",
        "GLuint64": "c_uint64",
        "GLuint64EXT": "c_uint64",
        "GLsync": "POINTER(struct___GLsync)",
        "GLDEBUGPROC": "CFUNCTYPE(None, GLenum, GLenum, GLuint, GLenum, GLsizei, POINTER(GLchar), POINTER(GLvoid))",
    }
    exclude_commands = set()

    def __init__(self, *, registry: Registry, out_file: Path):
        self._registry = registry
        self._out_file = out_file
        self._out = None
        self._all = []  # Entries for __all__

    def run(self):
        """Write the file and close"""
        self._out = open(self._out_file, mode='w')
        self.write_header()
        self.write_types()
        self.write_enums()
        self.write_commands()
        self.write_footer()
        self._out.close()

    def write(self, content: str):
        """Write out a string to the out file"""
        self._out.write(content)

    def write_lines(self, lines: str) -> None:
        """Write one or several lines to the out file."""
        for line in lines:
            self._out.write(line)
            self._out.write("\n")

    def write_header(self) -> None:
        """Write the header."""
        with open(REPO_ROOT / "tools" / "gl.template") as fd:
            self.write(fd.read())

    def write_types(self) -> None:
        """Write all types."""
        self.write_lines(["# GL type definitions"])
        self.write_lines([f"{k} = {v}" for k, v in self.types.items()])
        self.write_lines([""])
        self._all.extend(self.types.keys())

    def write_enums(self) -> None:
        """Write all enums."""
        self.write_lines(["# GL enumerant (token) definitions"])
        self.write_lines([
            f"{e.name} = {e.value_int}"
            for e in sorted(self._registry.enums.values())
        ])
        self.write_lines([""])
        self._all.extend(self._registry.enums.keys())

    def write_commands(self) -> None:
        """Write all commands."""
        self.write_lines(["# GL command definitions"])

        # _link_function params : name, restype, argtypes, requires=None, suggestions=None
        for cmd in sorted(self._registry.commands.values()):
            if cmd.name in self.exclude_commands:
                continue

            # Return type: If the function returns a pointer type ...
            if "*" in cmd.proto:
                restype = f"POINTER({cmd.ptype})"
            else:
                restype = cmd.ptype or "None"

            # Arguments can be pointer and pointer-pointer
            arguments = []
            for param in cmd.params:
                # print(cmd.name, param.name, param.ptype, "|", param.value)

                # Detect void pointers. They don't have a ptype set
                if "void" in param.value:
                    arguments.append("POINTER(GLvoid)")
                else:
                    # Ensure we actually know what the type is
                    if not self.types.get(param.ptype):
                        raise ValueError(f"ptype {param.ptype} not a known type")
                    # Handle pointer-pointer and pointers: *, **, *const*
                    if param.value.count("*") == 2:
                        arguments.append(f"POINTER(POINTER({param.ptype}))")
                    elif param.value.count("*") == 1:
                        arguments.append(f"POINTER({param.ptype})")
                    else:
                        arguments.append(param.ptype)

            argtypes = ", ".join(arguments)
            requires = f"OpenGL {cmd.requires}" if cmd.requires else "None"
            # NOTE: PROCs are optional
            # proc_name = f"PFN{cmd.name.upper()}PROC"

            self.write_lines([
                f"{cmd.name} = _link_function('{cmd.name}', {restype}, [{argtypes}], requires='{requires}')",
                # f"{proc_name} = CFUNCTYPE({restype}, {argtypes})",
            ])
            self._all.append(cmd.name)
            # self._all.append(proc_name)

        self.write_lines([""])

    def write_footer(self) -> None:
        """Write __all__ section."""
        self.write_lines([
            "",
            "__all__ = [",
            *[f"    '{name}'," for name in self._all],
            "]",
        ])
        

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
"""

__docformat__ = 'restructuredtext'
__version__ = '$Id$'

import pyglet

window = pyglet.window.Window()
image = pyglet.resource.image('kitten.jpg')


@window.event
def on_draw():
    window.clear()
    image.blit(0, 0)


pyglet.app.run()



#!/usr/bin/env python
"""
Usage

    media_player.py [options] <filename> [<filename> ...]

Plays the audio / video files listed as arguments, optionally
collecting debug info

Options
    --debug : saves sequence of internal state as a binary *.dbg
    --outfile : filename to store the debug info, defaults to filename.dbg

The raw data captured in the .dbg can be rendered as human readable
using the script report.py
"""


__docformat__ = 'restructuredtext'
__version__ = '$Id: $'

import os
import sys
import weakref

from pyglet.gl import (
    glEnable,
    glBlendFunc,
    GL_BLEND,
    GL_LINE_LOOP,
    GL_ONE_MINUS_SRC_ALPHA,
    GL_SRC_ALPHA,
)
import pyglet
from pyglet.window import key

pyglet.options['debug_media'] = False
# pyglet.options['audio'] = ('openal', 'pulse', 'silent')
from pyglet.media import buffered_logger as bl


def draw_rect(x, y, width, height, color=(1, 1, 1, 1)):
    pyglet.graphics.draw(
        4,
        GL_LINE_LOOP,
        position=('f', (x, y, 0,
                        x + width, y, 0,
                        x + width, y + height, 0,
                        x, y + height, 0,
                        ),
                  ),
        colors=('f', color * 4),
    )


class Control(pyglet.event.EventDispatcher):
    x = y = 0
    width = height = 10

    def __init__(self, parent):
        super(Control, self).__init__()
        self.parent = weakref.proxy(parent)

    def hit_test(self, x, y):
        return (self.x < x < self.x + self.width and
                self.y < y < self.y + self.height)

    def capture_events(self):
        self.parent.push_handlers(self)

    def release_events(self):
        self.parent.remove_handlers(self)


class Button(Control):
    charged = False

    def draw(self):
        if self.charged:
            draw_rect(self.x, self.y, self.width, self.height)
        else:
            draw_rect(self.x, self.y, self.width, self.height, color=(1, 0, 0, 1))
        self.draw_label()

    def on_mouse_press(self, x, y, button, modifiers):
        self.capture_events()
        self.charged = True

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        self.charged = self.hit_test(x, y)

    def on_mouse_release(self, x, y, button, modifiers):
        self.release_events()
        if self.hit_test(x, y):
            self.dispatch_event('on_press')
        self.charged = False


Button.register_event_type('on_press')


class TextButton(Button):
    def __init__(self, *args, **kwargs):
        super(TextButton, self).__init__(*args, **kwargs)
        self._text = pyglet.text.Label('', anchor_x='center', anchor_y='center')

    def draw_label(self):
        self._text.x = self.x + self.width / 2
        self._text.y = self.y + self.height / 2
        self._text.draw()

    def set_text(self, text):
        self._text.text = text

    text = property(lambda self: self._text.text,
                    set_text)


class Slider(Control):
    THUMB_WIDTH = 6
    THUMB_HEIGHT = 10
    GROOVE_HEIGHT = 2
    RESPONSIVNESS = 0.3

    def __init__(self, *args, **kwargs):
        super(Slider, self).__init__(*args, **kwargs)
        self.seek_value = None

    def draw(self):
        center_y = self.y + self.height / 2
        draw_rect(self.x, center_y - self.GROOVE_HEIGHT / 2,
                  self.width, self.GROOVE_HEIGHT)
        pos = self.x + self.value * self.width / (self.max - self.min)
        draw_rect(pos - self.THUMB_WIDTH / 2, center_y - self.THUMB_HEIGHT / 2,
                  self.THUMB_WIDTH, self.THUMB_HEIGHT)

    def coordinate_to_value(self, x):
        value = float(x - self.x) / self.width * (self.max - self.min) + self.min
        return value

    def on_mouse_press(self, x, y, button, modifiers):
        value = self.coordinate_to_value(x)
        self.capture_events()
        self.dispatch_event('on_begin_scroll')
        self.dispatch_event('on_change', value)
        pyglet.clock.schedule_once(self.seek_request, self.RESPONSIVNESS)

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        # On some platforms, on_mouse_drag is triggered with a high frequency.
        # Seeking takes some time (~200ms). Asking for a seek at every
        # on_mouse_drag event would starve the event loop.
        # Instead we only record the last mouse position and we
        # schedule seek_request to dispatch the on_change event in the future.
        # This will allow subsequent on_mouse_drag to change the seek_value
        # without triggering yet the on_change event.
        value = min(max(self.coordinate_to_value(x), self.min), self.max)
        if self.seek_value is None:
            # We have processed the last recorded mouse position.
            # We re-schedule seek_request
            pyglet.clock.schedule_once(self.seek_request, self.RESPONSIVNESS)
        self.seek_value = value

    def on_mouse_release(self, x, y, button, modifiers):
        self.release_events()
        self.dispatch_event('on_end_scroll')
        self.seek_value = None

    def seek_request(self, dt):
        if self.seek_value is not None:
            self.dispatch_event('on_change', self.seek_value)
            self.seek_value = None


Slider.register_event_type('on_begin_scroll')
Slider.register_event_type('on_end_scroll')
Slider.register_event_type('on_change')


class PlayerWindow(pyglet.window.Window):
    GUI_WIDTH = 400
    GUI_HEIGHT = 40
    GUI_PADDING = 4
    GUI_BUTTON_HEIGHT = 16

    def __init__(self, player):
        super(PlayerWindow, self).__init__(caption='Media Player',
                                           visible=False,
                                           resizable=True)
        # We only keep a weakref to player as we are about to push ourself
        # as a handler which would then create a circular reference between
        # player and window.
        self.player = weakref.proxy(player)
        self._player_playing = False
        self.player.push_handlers(self)

        self.slider = Slider(self)
        self.slider.push_handlers(self)
        self.slider.x = self.GUI_PADDING
        self.slider.y = self.GUI_PADDING * 2 + self.GUI_BUTTON_HEIGHT

        self.play_pause_button = TextButton(self)
        self.play_pause_button.x = self.GUI_PADDING
        self.play_pause_button.y = self.GUI_PADDING
        self.play_pause_button.height = self.GUI_BUTTON_HEIGHT
        self.play_pause_button.width = 45
        self.play_pause_button.on_press = self.on_play_pause

        self.window_button = TextButton(self)
        self.window_button.x = self.play_pause_button.x + \
                               self.play_pause_button.width + self.GUI_PADDING
        self.window_button.y = self.GUI_PADDING
        self.window_button.height = self.GUI_BUTTON_HEIGHT
        self.window_button.width = 90
        self.window_button.text = 'Windowed'
        self.window_button.on_press = lambda: self.set_fullscreen(False)

        self.controls = [
            self.slider,
            self.play_pause_button,
            self.window_button,
        ]

        x = self.window_button.x + self.window_button.width + self.GUI_PADDING
        i = 0
        for screen in self.display.get_screens():
            screen_button = TextButton(self)
            screen_button.x = x
            screen_button.y = self.GUI_PADDING
            screen_button.height = self.GUI_BUTTON_HEIGHT
            screen_button.width = 80
            screen_button.text = f'Screen {i + 1}'
            screen_button.on_press = lambda screen=screen: self.set_fullscreen(True, screen)
            self.controls.append(screen_button)
            i += 1
            x += screen_button.width + self.GUI_PADDING

        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def on_player_next_source(self):
        self.gui_update_state()
        self.gui_update_source()
        self.set_default_video_size()
        return True

    def on_player_eos(self):
        self.gui_update_state()
        pyglet.clock.schedule_once(self.auto_close, 0.1)
        return True

    def gui_update_source(self):
        if self.player.source:
            source = self.player.source
            self.slider.min = 0.
            self.slider.max = source.duration
        self.gui_update_state()

    def gui_update_state(self):
        if self.player.playing:
            self.play_pause_button.text = 'Pause'
        else:
            self.play_pause_button.text = 'Play'

    def get_video_size(self):
        if not self.player.source or not self.player.source.video_format:
            return 0, 0
        video_format = self.player.source.video_format
        width = video_format.width
        height = video_format.height
        if video_format.sample_aspect > 1:
            width *= video_format.sample_aspect
        elif video_format.sample_aspect < 1:
            height /= video_format.sample_aspect
        return width, height

    def set_default_video_size(self):
        """Make the window size just big enough to show the current
        video and the GUI."""
        width = self.GUI_WIDTH
        height = self.GUI_HEIGHT
        video_width, video_height = self.get_video_size()
        width = max(width, video_width)
        height += video_height
        self.set_size(int(width), int(height))

    def on_resize(self, width, height):
        """Position and size video image."""
        super(PlayerWindow, self).on_resize(width, height)
        self.slider.width = width - self.GUI_PADDING * 2

        height -= self.GUI_HEIGHT
        if height <= 0:
            return

        video_width, video_height = self.get_video_size()
        if video_width == 0 or video_height == 0:
            return
        display_aspect = width / float(height)
        video_aspect = video_width / float(video_height)
        if video_aspect > display_aspect:
            self.video_width = width
            self.video_height = width / video_aspect
        else:
            self.video_height = height
            self.video_width = height * video_aspect
        self.video_x = (width - self.video_width) / 2
        self.video_y = (height - self.video_height) / 2 + self.GUI_HEIGHT

    def on_mouse_press(self, x, y, button, modifiers):
        for control in self.controls:
            if control.hit_test(x, y):
                control.on_mouse_press(x, y, button, modifiers)

    def on_key_press(self, symbol, modifiers):
        if symbol == key.SPACE:
            self.on_play_pause()
        elif symbol == key.ESCAPE:
            self.dispatch_event('on_close')
        elif symbol == key.LEFT:
            self.player.seek(0)
        elif symbol == key.RIGHT:
            self.player.next_source()

    def on_close(self):
        self.player.pause()
        self.close()

    def auto_close(self, dt):
        self.close()

    def on_play_pause(self):
        if self.player.playing:
            self.player.pause()
        else:
            if self.player.time >= self.player.source.duration:
                self.player.seek(0)
            self.player.play()
        self.gui_update_state()

    def on_draw(self):
        self.clear()

        # Video
        if self.player.source and self.player.source.video_format:
            video_texture = self.player.texture
            video_texture.blit(self.video_x,
                               self.video_y,
                               width=self.video_width,
                               height=self.video_height)

        # GUI
        self.slider.value = self.player.time
        for control in self.controls:
            control.draw()

    def on_begin_scroll(self):
        self._player_playing = self.player.playing
        self.player.pause()

    def on_change(self, value):
        self.player.seek(value)

    def on_end_scroll(self):
        if self._player_playing:
            self.player.play()


def main(target, dbg_file, debug):
    set_logging_parameters(target, dbg_file, debug)

    player = pyglet.media.Player()
    window = PlayerWindow(player)

    player.queue(pyglet.media.load(filename) for filename in sys.argv[1:])

    window.gui_update_source()
    window.set_visible(True)
    window.set_default_video_size()

    # this is an async call
    player.play()
    window.gui_update_state()

    pyglet.app.run()


def set_logging_parameters(target_file, dbg_file, debug):
    if not debug:
        bl.logger = None
        return
    if dbg_file is None:
        dbg_file = target_file + ".dbg"
    else:
        dbg_dir = os.path.dirname(dbg_file)
        if dbg_dir and not os.path.isdir(dbg_dir):
            os.mkdir(dbg_dir)
    bl.logger = bl.BufferedLogger(dbg_file)
    from pyglet.media.instrumentation import mp_events
    # allow to detect crashes by prewriting a crash file, if no crash
    # it will be overwrited by the captured data
    sample = os.path.basename(target_file)
    bl.logger.log("version", mp_events["version"])
    bl.logger.log("crash", sample)
    bl.logger.save_log_entries_as_pickle()
    bl.logger.clear()
    # start the real capture data
    bl.logger.log("version", mp_events["version"])
    bl.logger.log("mp.im", sample)


def usage():
    print(__doc__)
    sys.exit(1)


def sysargs_to_mainargs():
    """builds main args from sys.argv"""
    if len(sys.argv) < 2:
        usage()
    debug = False
    dbg_file = None
    for i in range(2):
        if sys.argv[1].startswith("--"):
            a = sys.argv.pop(1)
            if a.startswith("--debug"):
                debug = True
            elif a.startswith("--outfile="):
                dbg_file = a[len("--outfile="):]
            else:
                print("Error unknown option:", a)
                usage()
    target_file = sys.argv[1]
    return target_file, dbg_file, debug


if __name__ == '__main__':
    target_file, dbg_file, debug = sysargs_to_mainargs()
    main(target_file, dbg_file, debug)
